#
# Copyright (C) 2023, Inria
# GRAPHDECO research group, https://team.inria.fr/graphdeco
# All rights reserved.
#
# This software is free for non-commercial, research and evaluation use
# under the terms of the LICENSE.md file.
#
# For inquiries contact  george.drettakis@inria.fr
#

import math
from .renderer import *
from diff_gaussian_rasterization import GaussianRasterizationSettings, GaussianRasterizer
from src.utils.sh import eval_sh


class VanillaRenderer(Renderer):
    def __init__(self, compute_cov3D_python: bool = False, convert_SHs_python: bool = False):
        super().__init__()

        self.compute_cov3D_python = compute_cov3D_python
        self.convert_SHs_python = convert_SHs_python

    def forward(
            self,
            viewpoint_camera: Camera,
            xyz,
            opacity,
            scales,
            rotations,
            shs,
            bg_color: torch.Tensor,
            colors_precomp = None,
            cov3D_precomp = None,
            scaling_modifier=1.0,
            override_color=None,
    ):
        """
        Render the scene.

        Background tensor (bg_color) must be on GPU!
        """
        # Create zero tensor. We will use it to make pytorch return gradients of the 2D (screen-space) means
        device = bg_color.device

        screenspace_points = torch.zeros_like(xyz, dtype=xyz.dtype, requires_grad=True, device=bg_color.device) + 0
        means3D = xyz                            
        means2D = screenspace_points
        try:
            screenspace_points.retain_grad()
        except:
            pass
        # Set up rasterization configuration
        tanfovx = math.tan(viewpoint_camera.fov_x * 0.5)
        tanfovy = math.tan(viewpoint_camera.fov_y * 0.5)

        raster_settings = GaussianRasterizationSettings(
            image_height=int(viewpoint_camera.height),
            image_width=int(viewpoint_camera.width),
            tanfovx=tanfovx,
            tanfovy=tanfovy,
            bg=bg_color,
            scale_modifier=scaling_modifier,
            viewmatrix=viewpoint_camera.world_to_camera.to(device),
            projmatrix=viewpoint_camera.full_projection.to(device),
            sh_degree=3,
            campos=viewpoint_camera.camera_center.to(device),
            prefiltered=False,
            debug=False
        )

        rasterizer = GaussianRasterizer(raster_settings=raster_settings)
        # Rasterize visible Gaussians to image, obtain their radii (on screen).

        dtype = xyz.dtype
        # to float32
        means3D = means3D.to(dtype=torch.float32)
        means2D = means2D.to(dtype=torch.float32)
        shs = shs.to(dtype=torch.float32)
        
        if len(shs.shape) == 2:
            colors_precomp = shs
            shs = None 

        scales = scales.to(dtype=torch.float32)
        rotations = rotations.to(dtype=torch.float32)
        opacity = opacity.to(dtype=torch.float32)
        if colors_precomp is not None:
            colors_precomp = colors_precomp.to(dtype=torch.float32)
        if cov3D_precomp is not None:
            cov3D_precomp = cov3D_precomp.to(dtype=torch.float32)

        rasterize_result = rasterizer(
            means3D=means3D,
            means2D=means2D,
            shs=shs,
            colors_precomp=colors_precomp,
            opacities=opacity,
            scales=scales,
            rotations=rotations,
            cov3D_precomp=cov3D_precomp,
        )

        depth_image = None
        alpha_image = None
        rendered_image = None
        radii = None 
        if len(rasterize_result) == 2:
            rendered_image, radii = rasterize_result
        if len(rasterize_result) == 3:
            rendered_image, radii, depth_image = rasterize_result
        if len(rasterize_result) == 4:
            rendered_image, radii, depth_image, alpha_image = rasterize_result

        # Those Gaussians that were frustum culled or had a radius of 0 were not visible.
        # They will be excluded from value updates used in the splitting criteria.

        return {
            "render": rendered_image.to(dtype=dtype),
            "depth": depth_image.to(dtype=dtype) if depth_image is not None else None,
            # "viewspace_points": screenspace_points,
            # "visibility_filter": radii > 0,
            # "radii": radii,
            "means3D": means3D.to(dtype=dtype),
            "scale": scales.to(dtype=dtype),
            "rotation": rotations.to(dtype=dtype),
            "opacity": opacity.to(dtype=dtype),
            # "color": colors_precomp,
        }
